PMAP = {
    "C": "Basic (code)", "I": "Implicit (id_token)",
    "IT": "Implicit (id_token+token)",
    "CI": "Hybrid (code+id_token)", "CT": "Hybrid (code+token)",
    "CIT": "Hybrid (code+id_token+token)"
}
